/**
 * PithHelmet
 * @author Mike Solomon
 * Copyright 2003 Mike Solomon. All rights reserved.
 * $Header: /Users/msolomon/Developer/cvs/PithHelmet/PCRegExp.m,v 1.8 2003/11/12 07:38:41 msolomon Exp $
 */

RCS_ID("$Id: PCRegExp.m,v 1.8 2003/11/12 07:38:41 msolomon Exp $")

#import "PCRegExp.h"
#import "pcre.h"

@implementation PCRegExp

+ (BOOL) validatePattern:(NSString*)_pattern errorIndex:(int*)_errorIndex errorMessage:(NSString**)_errorMessage
{
	const char* errorString;	
	
	pcre* tempRegex = kjs_pcre_compile([_pattern cString], PCRE_CASELESS, &errorString, _errorIndex, NULL);
	if (!tempRegex)
	{
		if (errorString)
			*_errorMessage = [NSString stringWithCString:errorString];

		DLog1(DLOG_10, (@"regex validation error at index %d : %@", *_errorIndex, *_errorMessage));
		return NO;
	}
		
	pcre_extra* tempRegexExtra = kjs_pcre_study(tempRegex, 0x0, &errorString);
	if (!tempRegexExtra && errorString)
	{
		*_errorMessage = [NSString stringWithCString:errorString];

		DLog1(DLOG_10, (@"regex extras validation error at index %d : %@", *_errorIndex, *_errorMessage));
		return NO;
	}

	kjs_pcre_free(tempRegex);
	kjs_pcre_free(tempRegexExtra);

	return YES;
}


- (id)initWithString:(NSString*)_pattern
{
	pattern = nil;
	regex = nil;
	regex_extra = nil;
	
	self = [super init];
	if (self == nil)
		return nil;
	
	if ([self setPattern:_pattern] == NO)
	{
		[self dealloc];
		return nil;
	}
	return self;
}

- (void) dealloc
{
	if (pattern)
		[pattern release];
	pattern = nil;
	
	if (regex)
		kjs_pcre_free(regex);
	regex = nil;
	
	if (regex_extra)
		kjs_pcre_free(regex_extra);
	regex_extra = nil;
	
	[super dealloc];
}

/**
 * Reset the pattern we are matching - this is semi-expensive.
 * @return NO if there is an error
 */
- (BOOL)setPattern:(NSString*)_pattern
{
	const char* errstr;
	int error_offset; // offset into the substring where the error occurred
	
	// default this object to invalid until the pattern compiles
	valid = NO;
	
	if (pattern)
		[pattern autorelease];
	
	// copy in case the original was mutable and changes
	pattern = [_pattern copy];
	
	if (regex)
		kjs_pcre_free(regex);
	
	regex = kjs_pcre_compile([_pattern cString], PCRE_CASELESS, &errstr, &error_offset, NULL);
	if (!regex)
	{
		NSLog(@"compilation of regex(%@) failed at: %d with %s", pattern, error_offset, errstr);
		return NO;
	}
	if (regex_extra)
		kjs_pcre_free(regex_extra);
		
	regex_extra = kjs_pcre_study(regex, 0x0, &errstr);
	if (!regex_extra && errstr)
	{
		NSLog(@"study of regex(%@) failed with %s", pattern, errstr);
		return NO;
	}
	
	valid = YES;
	return YES;
}

- (NSString*)pattern
{
	return pattern;
}

- (BOOL) isValid
{
	return valid;
}

- (BOOL)matchInString:(NSString*)_string
{
	int n;
	
	n = kjs_pcre_exec(regex, regex_extra, [_string cString], [_string cStringLength], 0, 0, offsetArray, PCREGEXP_MAX_SUB_PATTERNS);
	
	if (n == PCRE_ERROR_NOMATCH)
		return NO;
	
	// 0 indicates that offsetArray was too small, but we don't care
	if (n < 0)
	{
		NSLog(@"pcre_exec error = %d", n);
		return NO;
	}

	return YES;
}

/*
 * returns # of subs made
 *
 * this fxn copied almost verbatim from Max Okumoto's code.
 */
- (int) substituteFrom:(NSString*) _srcString intoString:(NSMutableString*) _dstString usingPattern:(NSString*) _replacementPattern andMatch:(int) nmatch
{
	int nsubs = 0;
	const char *p = [_replacementPattern cString];
	int subnum = 0;
	int state = 0;
	
	while (state != -1)
	{
		switch (state)
		{
			case 0:
				if (!*p)
				{
					state = -1;
					break;
				}
				if (*p == '$')
				{
					state = 1;
					subnum = 0;
					if (p[1] == '&')
					{
						p++;
						if (isdigit(p[1]))
							p++;
					}
					else if (!isdigit(p[1]))
					{
						NSLog(@"badly formed replacement pattern: %@", _replacementPattern);
						return -1;
					}
				}
				else 
					[_dstString appendFormat:@"%c", *p];
				break;
				
			case 1:
				if (isdigit(*p))
				{
					subnum *= 10;
					subnum += (*p) - '0';
				}
				else
				{
					char substr[4096];
					int status;
					
					status = kjs_pcre_copy_substring([_srcString cString], offsetArray, nmatch, subnum, substr, 4096);
					
					switch (status)
					{
						case PCRE_ERROR_NOMEMORY:	/* buffer too small */
						case PCRE_ERROR_NOSUBSTRING:	/* bad stringnumber */
							NSLog(@"substitution error %d", status);
							return -1;
					}
					[_dstString appendFormat: @"%s", substr];
					nsubs++;
					
					state = 0;
					continue;	/* send char to start state */
				}
		}
		++p;
	}
	return nsubs;
}

/*
 * substitute
 *  returns number of substitutions made
 */
- (NSString*) replaceString:(NSString*)_string byPattern:(NSString*)_pattern
{
	int nmatch;
	int subsitutionCount = 0;
	NSMutableString* final = [NSMutableString stringWithCapacity: 256];
	
	int offset = 0;
	int last = 0;
	while (true)
	{
		// find next matching subs
		nmatch = kjs_pcre_exec(regex, regex_extra, [_string cString], [_string cStringLength], offset, 0, offsetArray, PCREGEXP_MAX_SUB_PATTERNS);
	
		// doesn't work exactly like perl
		if (nmatch <= 0)
		{
			if (nmatch != PCRE_ERROR_NOMATCH)
			{
				NSLog(@"pcre_exec error = %d", nmatch);
				return nil;
			}
			else
			{
				break;
			}
		}
		
		// append anything previously unmatched, but not substituted
		if (last <= offsetArray[0])
		{
			[final appendString:[_string substringWithRange: NSMakeRange(last, offsetArray[0] - last)]];
			last = offsetArray[1];
		}
		
		// replace stuff in s
		subsitutionCount += [self substituteFrom:_string intoString:final usingPattern:_pattern andMatch:nmatch];
		//subsitutionCount += do_subs(_string, final, _pattern, nmatch);
		
		// if global gotta check match at every pos
		//if (!(_pcre_opts & PREGEX::_PREGEX_GLOBAL))
		//	break;
		
		if (offsetArray[0] != offsetArray[1])
			offset = offsetArray[1];
		else
		{
			// matched empty string
			if (offsetArray[1] == [_string length])
				break;
				
			offset = offsetArray[1] + 1;
		}
	}
//	delete[] ovec;
	[final appendString: [_string substringFromIndex: last]];
	
	return final;
}


@end
