/**
 * PithHelmet
 * Copyright 2003, Mike Solomon <mas63@cornell.edu>
 * $Header: /Users/msolomon/Developer/cvs/PithHelmet/PithHelmet.m,v 1.55 2003/11/13 17:29:04 msolomon Exp $
 */

RCS_ID("$Id: PithHelmet.m,v 1.55 2003/11/13 17:29:04 msolomon Exp $")

#import "PithHelmet.h"
#import "PithHelmetBrowserWebView.h"
#import "PithHelmetRuleNode.h"
#import "PithHelmetWebImageRenderer.h"
#import "PithHelmetLocationChangeHandler.h"
#import "PithHelmetLoadProgressMonitor.h"
#import "NSPreferences.h"
#import "PithHelmetPreferences.h"

@implementation PithHelmet

+ (void)load
{
	DLog1(DLOG_50, (@"loaded PithHelmet"));
	
	NS_DURING
		int safariVersion = [[[NSBundle mainBundle] objectForInfoDictionaryKey:@"CFBundleVersion"] intValue];
		int minVersion = [[[NSBundle bundleWithIdentifier:PH_IDENTIFIER] objectForInfoDictionaryKey:PH_INFO_KEY_MIN_SAFARI_BUNDLE_VERSION] intValue];
		int maxVersion = [[[NSBundle bundleWithIdentifier:PH_IDENTIFIER] objectForInfoDictionaryKey:PH_INFO_KEY_MAX_SAFARI_BUNDLE_VERSION] intValue];
		if (safariVersion > maxVersion || safariVersion < minVersion)
			[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Safari v%d has not been tested with this version of PithHelmet.", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error alert primary message"), safariVersion];
		
		PithHelmet* pithHelmet = [PithHelmet sharedInstance];
		[pithHelmet readUserDefaults];
		[pithHelmet checkFiles];
		
		NSString* userPlistPath = [PH_PLIST_PATH stringByExpandingTildeInPath];
		NSDictionary* userPlistDict = [pithHelmet readRulesFromFile:userPlistPath asMutable:NO];
		NSString* bundlePlistPath = [[NSBundle bundleWithIdentifier:PH_IDENTIFIER] pathForResource:@"PithHelmet" ofType:@"plist"];
		NSDictionary* bundlePlistDict = [pithHelmet readRulesFromFile:bundlePlistPath asMutable:NO];
		
		int userVersion = [userPlistDict intForKey:PH_PLIST_KEY_VERSION];
		int bundleVersion = [bundlePlistDict intForKey:PH_PLIST_KEY_VERSION];
		// if this is an outdated version, lets make a backup copy of the file
		// and copy the defaults in to replace them.
		if (userVersion < bundleVersion)
		{
			[pithHelmet makeBackupFile:userPlistPath];
			// fixme: this shouldn't leak, also it only gets exec'd once
			userPlistDict = [pithHelmet mergeFactoryPlist:bundlePlistDict intoUserPlist:userPlistDict];
			[pithHelmet saveRulesDictionary:userPlistDict toFile:userPlistPath];
		}
		
		[pithHelmet loadRulesFromDictionary:userPlistDict];
	
		DLog1(DLOG_10, (@"adding NSApplicationWillTerminateNotification"));
		// add a handler to save and close the log file when the app quits
		[[NSNotificationCenter defaultCenter] addObserver:pithHelmet selector:@selector(appWillTerminate:) name:NSApplicationWillTerminateNotification object:nil];
	
		// start intercepting calls for various web resources
		DLog1(DLOG_10, (@"adding class overrides"));

		if (NSClassFromString(@"LocationChangeHandler") == nil)
			[NSException raise:@"PithHelmet Error" format:@"Failed loading class LocationChangeHandler"];
		[[PithHelmetLocationChangeHandler class] poseAsClass: [LocationChangeHandler class]];

		if (NSClassFromString(@"WebImageRenderer") == nil)
			[NSException raise:@"PithHelmet Error" format:@"Failed loading class WebImageRenderer"];
		[[PithHelmetWebImageRenderer class] poseAsClass: [WebImageRenderer class]];

		if (NSClassFromString(@"LoadProgressMonitor") == nil)
			[NSException raise:@"PithHelmet Error" format:@"Failed loading class LoadProgressMonitor"];
		[[PithHelmetLoadProgressMonitor class] poseAsClass: [LoadProgressMonitor class]];

		if (NSClassFromString(@"BrowserWebView") == nil)
			[NSException raise:@"PithHelmet Error" format:@"Failed loading class BrowserWebView"];
		[[PithHelmetBrowserWebView class] poseAsClass: [BrowserWebView class]];

		// this pre-loads the prefs pane and inserts the menu item overrides
		[NSBundle loadNibNamed:[pithHelmet nibName] owner:[PithHelmetPreferences sharedInstance]];

		// everything looks good, so we need to let the prefs pane know somehow
		[pithHelmet setLoaded:YES];
	NS_HANDLER
		NSLog(@"%@: %@", [localException name], [localException reason]);
		[PithHelmet errorAlert:NSLocalizedStringFromTableInBundle(@"An error occurred loading PithHelmet.", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error alert primary message") withDetails:[localException reason]];
	NS_ENDHANDLER
}

/**
 * @return the single static instance of the PithHelmet object
 */
+ (PithHelmet*)sharedInstance
{
	static PithHelmet* pithHelmet = nil;
 	
	if (pithHelmet == nil)
		pithHelmet = [[PithHelmet alloc] init];
		
	return pithHelmet;
}


/**
 * Recursively create a directory path.
 * @throws exception on failure creating the directories
 */
- (void) createDirectoryAtPath:(NSString*)_directoryPath
{
	NSFileManager* fileManager = [NSFileManager defaultManager];
	if ([fileManager fileExistsAtPath:_directoryPath] == YES)
		return;
	
	NSString* parentDirectory = [_directoryPath stringByDeletingLastPathComponent];
	if ([fileManager fileExistsAtPath:parentDirectory] == YES)
	{
		if ([fileManager createDirectoryAtPath:_directoryPath attributes:nil] == NO)
			[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed creating directory %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute the directory path"), _directoryPath];
	}
	else
	{
		[self createDirectoryAtPath:parentDirectory];
	}
}


/**
 * Checks the PithHelmet files, creating them if necessary before loads are attempted.
 * @throws exception on file copy/create errors
 */
- (void) checkFiles
{
	NSString* plistPath = [PH_PLIST_PATH stringByExpandingTildeInPath];
	DLog1(DLOG_10, (@"checking path %@", plistPath));
	
	NSFileManager* fileManager = [NSFileManager defaultManager];
	if ([fileManager fileExistsAtPath:plistPath] == NO)
	{
		NSString* bundlePlistPath = [[NSBundle bundleWithIdentifier:PH_IDENTIFIER] pathForResource:@"PithHelmet" ofType:@"plist"];
		if ([fileManager fileExistsAtPath:bundlePlistPath] == YES)
		{
			[self createDirectoryAtPath:[plistPath stringByDeletingLastPathComponent]];
			if ([fileManager copyPath:bundlePlistPath toPath:plistPath handler:nil] == NO)
				[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed copying default rule set %@ to %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute the directory paths"), bundlePlistPath, plistPath];
		}
	}
	
	if ([[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_ENABLE_LOGGING] == YES)
	{
		NSString* logPath = [PH_LOG_PATH stringByExpandingTildeInPath];
		if ([fileManager fileExistsAtPath:logPath] == NO)
			if ([fileManager createFileAtPath:logPath contents:nil attributes:nil] == NO)
				[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed to create file %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute the file path"), logPath];

		[self loadLogFile];
	}
}

/**
 * @return either a NSDictionary or NSMutableDictionary depending on the _mutable param
 * @throws exception on failure reading file
 */
- (id) readRulesFromFile:(NSString*)_filePath asMutable:(BOOL)_mutable
{
	DLog1(DLOG_10, (@"readRulesFromFile %@", _filePath));

	NSData* plistData = [NSData dataWithContentsOfFile:_filePath];
	if (plistData == nil)
		[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed loading data from %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute the directory path"), _filePath];

	NSString* error;
	NSPropertyListFormat format;
	id plistDict = [NSPropertyListSerialization propertyListFromData:plistData mutabilityOption:(_mutable ? NSPropertyListMutableContainers : NSPropertyListImmutable) format:&format errorDescription:&error];
	
	if (plistDict == nil)
		[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed deserializing plist: %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute raw error message details"), error];
	
	return plistDict;
}

/**
 * Takes _plistArray (array of rule dictionaries) and creates rule objects placed
 * into _rulesArray.
 */
+ (void) populateRuleArray:(NSMutableArray*)_ruleArray fromPlistArray:(NSArray*)_plistArray
{
	if (_plistArray == nil)
		NSLog(@"Nil rule array - this is unusual");  // fixme: examine this for related error conditions
	else
	{
		NSDictionary* ruleDictionary;
	
		NSEnumerator* ruleEnumerator = [_plistArray objectEnumerator];
		while (ruleDictionary = [ruleEnumerator nextObject])
		{
			DLog1(DLOG_30, (@"adding rule %@", ruleDictionary));
			PithHelmetRuleNodeData* ruleData = [PithHelmetRuleNodeData ruleDataWithDictionary:ruleDictionary];
			if (ruleData != nil)
				[_ruleArray addObject:ruleData];
			else
				NSLog(@"skipping rule %@", ruleDictionary);
		}
	}
}

/**
 * Turns the rules plist into actual rule objects
 * @throws exception on failure reading rules, but not on syntax errors in the rules
 */
- (void) loadRulesFromDictionary:(NSDictionary*)_rulesPlist
{
	[PithHelmet populateRuleArray:blockRules fromPlistArray:[_rulesPlist objectForKey:PH_PLIST_KEY_BLOCK]];
	[PithHelmet populateRuleArray:allowRules fromPlistArray:[_rulesPlist objectForKey:PH_PLIST_KEY_ALLOW]];
	[PithHelmet populateRuleArray:exemptSiteRules fromPlistArray:[_rulesPlist objectForKey:PH_PLIST_KEY_EXEMPT_SITES]];
	[PithHelmet populateRuleArray:transmogrificationRules fromPlistArray:[_rulesPlist objectForKey:PH_PLIST_KEY_TRANSMOGRIFICATIONS]];

	[self buildContentScriptlet];
}

/**
 * Arrays are rule dictionary objects
 * @returns merged array - new factory rule + user rules
 */
+ (NSArray*) mergeFactoryRules:(NSArray*)_factoryRules intoUserRules:(NSArray*)_userRules
{
	if (_userRules == nil || [_userRules count] == 0)
		return _factoryRules;

	NSMutableArray* mergedRules = [NSMutableArray arrayWithCapacity:[_factoryRules count] + [_userRules count]];
	
	// clone over factory rules
	[mergedRules addObjectsFromArray:_factoryRules];

	// append non-factory rules from the user
	int i;
	for (i = [_userRules count] - 1; i >= 0; i--)
	{
		NSDictionary* rule = [_userRules objectAtIndex:i];
		if ([rule boolForKey:PH_RULE_KEY_FACTORY_RULE] == NO)
			[mergedRules addObject:rule];
	}

	return mergedRules;
}

/**
 * Reads the rules plist inside the bundle and merges the factory-defined rules into
 * the given plist file.
 */
- (NSDictionary*) mergeFactoryPlist:(NSDictionary*)_factoryPlist intoUserPlist:(NSDictionary*)_userPlist
{
	// contrary to what it says above, merge the user rules into the factory rules - 
	// this allows us to inherit everything from the factory file by default

	NSArray* userRulesArray;
	NSMutableDictionary* mergedPlist = [NSMutableDictionary dictionaryWithDictionary:_factoryPlist];

	// too old to use factory-rule merging code
	if ([_userPlist intForKey:PH_PLIST_KEY_VERSION] < 3)
		return mergedPlist;

	// merge allow, block and exempt sites rules
	userRulesArray = [PithHelmet mergeFactoryRules:[_factoryPlist objectForKey:PH_PLIST_KEY_ALLOW] intoUserRules:[_userPlist objectForKey:PH_PLIST_KEY_ALLOW]];
	[mergedPlist setObject:userRulesArray forKey:PH_PLIST_KEY_ALLOW];

	userRulesArray = [PithHelmet mergeFactoryRules:[_factoryPlist objectForKey:PH_PLIST_KEY_BLOCK] intoUserRules:[_userPlist objectForKey:PH_PLIST_KEY_BLOCK]];
	[mergedPlist setObject:userRulesArray forKey:PH_PLIST_KEY_BLOCK];

	userRulesArray = [PithHelmet mergeFactoryRules:[_factoryPlist objectForKey:PH_PLIST_KEY_EXEMPT_SITES] intoUserRules:[_userPlist objectForKey:PH_PLIST_KEY_EXEMPT_SITES]];
	[mergedPlist setObject:userRulesArray forKey:PH_PLIST_KEY_EXEMPT_SITES];

	// prior to version 6, domain and url rules were in separate keys
	if ([_userPlist intForKey:PH_PLIST_KEY_VERSION] < 6)
	{
		userRulesArray = [PithHelmet mergeFactoryRules:[mergedPlist objectForKey:PH_PLIST_KEY_ALLOW] intoUserRules:[_userPlist objectForKey:PH_PLIST_KEY_ALLOW_DOMAINS]];
		[mergedPlist setObject:userRulesArray forKey:PH_PLIST_KEY_ALLOW];
		userRulesArray = [PithHelmet mergeFactoryRules:[mergedPlist objectForKey:PH_PLIST_KEY_BLOCK] intoUserRules:[_userPlist objectForKey:PH_PLIST_KEY_BLOCK_DOMAINS]];
		[mergedPlist setObject:userRulesArray forKey:PH_PLIST_KEY_BLOCK];
	}

	return mergedPlist;
}

/**
 * Renames _filePath by appending .bak.n where n is the next logical number in a sequence
 * @throws exception on failure reading factory file or saving merged file
 */
- (void) makeBackupFile:(NSString*)_filePath
{
	unsigned maxBackupNumber = 0;
	
	NSFileManager* fileManager = [NSFileManager defaultManager];
	NSString* parentDirectory = [_filePath stringByDeletingLastPathComponent];
	NSEnumerator* dirListing = [[fileManager directoryContentsAtPath:parentDirectory] objectEnumerator];

	NSString* fileName = [_filePath lastPathComponent];
	NSString* fileNamePrefix = [NSString stringWithFormat:@"%@.bak.", fileName];
	
	NSString* dirEntry;
	while (dirEntry = [dirListing nextObject])
	{
		if ([dirEntry hasPrefix:fileNamePrefix] == YES)
		{
			int backupNumber = [[dirEntry pathExtension] intValue];
			if (backupNumber > maxBackupNumber)
				maxBackupNumber = backupNumber;
		}
	}
	maxBackupNumber++;
	
	NSString* backupFilePath = [parentDirectory stringByAppendingPathComponent:[NSString stringWithFormat:@"%@%d", fileNamePrefix, maxBackupNumber]];
	if ([fileManager movePath:_filePath toPath:backupFilePath handler:nil] != YES)
		[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed copying %@ to %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute file paths"), _filePath, backupFilePath];
}


/**
 * Checks that file exists and creates the parent directories if specified, and if necessary.
 * @return YES if file exist, NO otherwise
 * @throws exception on failure creating the directories
 */
- (BOOL) fileExistsAtPath:(NSString*)_filePath createDirectories:(BOOL)_createDirectories
{
	NSFileManager* fileManager = [NSFileManager defaultManager];
	
	BOOL fileExists = [fileManager fileExistsAtPath:_filePath];
	
	if (fileExists == NO && _createDirectories == YES)
		[self createDirectoryAtPath:[_filePath stringByDeletingLastPathComponent]];
		
	return fileExists;
}


/**
 * Opens a mutable copy of the user plist, overwrites the current rules and saves it to the
 * specified file.
 * @throws exception on failure reading or writing the file
 */
- (void) saveRulesToFile:(NSString*)_filePath
{
	NSMutableDictionary* pldict = [self readRulesFromFile:_filePath asMutable:YES];
	
	NSMutableArray* tmpBlockRules = [NSMutableArray array];
	NSMutableArray* tmpAllowRules = [NSMutableArray array];
	NSMutableArray* tmpExemptSiteRules = [NSMutableArray array];
	NSMutableArray* tmpTransmogRules = [NSMutableArray array];
	
	[pldict setObject:tmpAllowRules forKey:PH_PLIST_KEY_ALLOW];
	[pldict setObject:tmpBlockRules forKey:PH_PLIST_KEY_BLOCK];
	[pldict setObject:tmpExemptSiteRules forKey:PH_PLIST_KEY_EXEMPT_SITES];
	[pldict setObject:tmpTransmogRules forKey:PH_PLIST_KEY_TRANSMOGRIFICATIONS];
	
	NSEnumerator* e;
	PithHelmetRuleNodeData* ruleData;
	
	e = [allowRules objectEnumerator];
	while (ruleData = [e nextObject])
		[tmpAllowRules addObject:[ruleData asDictionary]];

	e = [blockRules objectEnumerator];
	while (ruleData = [e nextObject])
		[tmpBlockRules addObject:[ruleData asDictionary]];

	e = [exemptSiteRules objectEnumerator];
	while (ruleData = [e nextObject])
		[tmpExemptSiteRules addObject:[ruleData asDictionary]];

	e = [transmogrificationRules objectEnumerator];
	while (ruleData = [e nextObject])
		[tmpTransmogRules addObject:[ruleData asDictionary]];
	
	[self saveRulesDictionary:pldict toFile:_filePath];
}

/**
 * Write out the rules dictionary to the specified file.
 * @throws exception on failure writing the file
 */
- (void) saveRulesDictionary:(NSDictionary*)_rulesPlist toFile:(NSString*)_filePath
{
	DLog1(DLOG_20, (@"saveRulesDictionary:\n%@", _rulesPlist));
	
	NSString* errorDescription;
	NSData* data = [NSPropertyListSerialization dataFromPropertyList:_rulesPlist format:NSPropertyListXMLFormat_v1_0 errorDescription:&errorDescription];
	if (data == nil)
		[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed serializing dictionary: %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute raw error message details"), errorDescription];
	
	// NOTE: if you write atomically, symlinks are ignored - not so good - an understandable bug
	// the "mv" command doesn't respect symlinks either.
	if ([data writeToFile:_filePath atomically:NO] != YES)
		[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed to create file %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute the file path"), _filePath];
}

-(void) toggleContentFiltering
{
	NSUserDefaults* defaults =[NSUserDefaults standardUserDefaults];
	[defaults setBool:![defaults boolForKey:PH_PREF_KEY_ENABLE_FILTERING] forKey:PH_PREF_KEY_ENABLE_FILTERING];
}


-(void) reloadRules
{
	[blockRules removeAllObjects];
	[allowRules removeAllObjects];
	[exemptSiteRules removeAllObjects];
	[transmogrificationRules removeAllObjects];

	NSDictionary* userPlistDict = [self readRulesFromFile:[PH_PLIST_PATH stringByExpandingTildeInPath] asMutable:NO];
	[self loadRulesFromDictionary:userPlistDict];
}

/**
 * Mulch user rules in favor of just the current default rules.
 * Also reloads the active rule set.
 */
- (void) revertToFactoryRules
{
	NSString* userPlistPath = [PH_PLIST_PATH stringByExpandingTildeInPath];
	NSString* bundlePlistPath = [[NSBundle bundleWithIdentifier:PH_IDENTIFIER] pathForResource:@"PithHelmet" ofType:@"plist"];
	NSDictionary* bundlePlistDict = [self readRulesFromFile:bundlePlistPath asMutable:NO];
		
	[self makeBackupFile:userPlistPath];
	[self saveRulesDictionary:bundlePlistDict toFile:userPlistPath];		
	[self loadRulesFromDictionary:bundlePlistDict];
}

/**
 * Reloads the current url (frontmost browser document) without any of PithHelmet's functionality.
 */
-(void) reloadWithoutFilter
{
	id browserDocument = [[BrowserDocumentController sharedDocumentController] frontmostBrowserDocument];
	DLog1(DLOG_40, (@"reloadWithoutFilter:%@", [browserDocument currentURL]));
	[self startLoadWithoutFilter:[browserDocument currentURL]];
	[browserDocument clearPageCache];
	[browserDocument reload:nil];
}

- (void) startLoadWithoutFilter:(NSURL*)_url
{
	DLog1(DLOG_40, (@"startLoadWithoutFilter:<%@>\"%@\"", [_url class], _url));
	[reloadUrlsWithoutFilter setObject:[NSNumber numberWithBool:YES] forKey:[_url absoluteString]];
}

- (BOOL) shouldLoadWithoutFilter:(NSURL*)_url
{
	NSString* scheme = [_url scheme];
	NSString* host = [_url host];

	if ([scheme isEqualToString:@"file"] || [host isEqualToString:@"localhost"] || [host isEqualToString:@"127.0.0.1"])
		return YES;

	PithHelmetRuleNodeData* exemptSiteRule;
	NSEnumerator* exemptSiteEnumerator = [[self exemptSiteRules] objectEnumerator];
		
	// fixme: maybe if we match, we should dump this into the hashtable to prevent rescanning all the regular expressions?
	// this will get hit multiple times per page load...
	// if we match an include pattern explicitly, continue loading
	while (exemptSiteRule = [exemptSiteEnumerator nextObject])
	{
		if ([exemptSiteRule matchURL:_url] == YES)
			return YES;
	}

	id obj = [reloadUrlsWithoutFilter objectForKey:[_url absoluteString]];
	DLog1(DLOG_40, (@"PH shouldLoadWithoutFilter:<%@>\"%@\" (%d) (%@)", [_url class], _url, (obj != nil), obj));
	DLog1(DLOG_30, (@"PH check reloadUrlsWithoutFilter:%@", reloadUrlsWithoutFilter));
	return (obj != nil);
}

- (void) finishLoad:(NSURL*)_url
{
	DLog1(DLOG_40, (@"PH finishLoad: \"%@\"", _url));
	DLog1(DLOG_30, (@"PH clear reloadUrlsWithoutFilter:%@", reloadUrlsWithoutFilter));
	[reloadUrlsWithoutFilter removeObjectForKey:[_url absoluteString]];
}


- (id) init
{	
	totalUrlCount = 0;
	blockedUrlCount = 0;
	allowedUrlCount = 0;
	
	collapseContentScriptlet = nil;
	
	// fixme: this seems a little dopey - maybe the other methods should just call set?
	// of course, all the code currently assumes these values will be non-nil
	blockRules = [[NSMutableArray alloc] init];
	allowRules = [[NSMutableArray alloc] init];
	exemptSiteRules = [[NSMutableArray alloc] init];
	transmogrificationRules = [[NSMutableArray alloc] init];
	reloadUrlsWithoutFilter = [[NSMutableDictionary alloc] init];
	
	loaded = NO;

	return self;
}

- (void) appWillTerminate:(NSNotification*)_notification
{
	[self saveUserDefaults];

	if (logFileHandle == nil)
		return;
	
	[logFileHandle synchronizeFile];
	[logFileHandle closeFile];
}

/**
 * @throws exception if file can't be opened for writing
 */
- (void)loadLogFile
{
	NSString* logPath = [PH_LOG_PATH stringByExpandingTildeInPath];


	logFileHandle = [NSFileHandle fileHandleForWritingAtPath:logPath];
	if (logFileHandle == nil)
		[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed to open file %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute the file path"), logPath];
		
	[logFileHandle retain];
	[logFileHandle seekToEndOfFile];
}

- (void)truncateLogFile
{
	if (logFileHandle != nil)
		{
			[logFileHandle synchronizeFile];
			[logFileHandle closeFile];
			[logFileHandle autorelease];
			logFileHandle = nil;
		}

	NSString* logPath = [PH_LOG_PATH stringByExpandingTildeInPath];
	NSFileManager* fileManager = [NSFileManager defaultManager];

	if ([fileManager fileExistsAtPath:logPath] == YES && [fileManager removeFileAtPath:logPath handler:nil] == NO)
		[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed to delete file %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute the file path"), logPath];

	if ([fileManager createFileAtPath:logPath contents:nil attributes:nil] == NO)
		[NSException raise:@"PithHelmet Error" format:NSLocalizedStringFromTableInBundle(@"Failed to create file %@", PH_STRING_TABLE, [NSBundle bundleWithIdentifier:PH_IDENTIFIER], @"Error message, substitute the file path"), logPath];
}

/**
 * Create a quick tab delimited entry in the PithHelmet log
 * @param _action either "allow", "block", "transmogrify"
 * @param _testMode if yes, returns a handy log string for the PithHelmet test string
 * @return the log string
 */
- (NSString*)logAction:(NSString*)_action forUrl:(NSString*)_url withReferrer:(NSString*)_referrer byRule:(NSString*)_rule withResult:(NSString*)_result testMode:(BOOL)_testMode;
{
	if (_testMode == NO &&
		[[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_ENABLE_LOGGING] == NO)
		return nil;
	
	if (_testMode == YES)
	{
		NSString* logEntry = [NSString stringWithFormat:@"action: %@\nrule: %@\nresult: %@\n", _action, _rule, _result == nil ? @"none" : _result];
		DLog1(DLOG_10, (@"logAction: %@", logEntry));
		return logEntry;
	}

	NSString* logEntry = [NSString stringWithFormat:@"%@\t%@\t%@\t%@\t%@\t%lu\n", _action, _rule, _url, _referrer == nil ? @"" : _referrer , _result == nil ? @"" : _result, (unsigned long)[[NSDate date] timeIntervalSince1970]];
	[logFileHandle writeData:[logEntry dataUsingEncoding:NSUTF8StringEncoding]];
	return logEntry;
}


- (BOOL) allowResourceLoad:(NSURL*)_url withReferrer:(NSURL*)_referrer 
{
	NSString* logResults;
	return [self allowResourceLoad:_url withReferrer:_referrer testMode:NO logResults:&logResults];
}

/**
 * This is the basic brain here. Logic is thusly:
 * 1. are we a sub-resource? check cross-site rule if enabled
 * 2. check block rules - if we hit, examine allow rules
 */
- (BOOL) allowResourceLoad:(NSURL*)_url withReferrer:(NSURL*)_referrer testMode:(BOOL)_testMode logResults:(NSString**)_logResults
{
	// this is a little bit of a cheat, since this would also increment if you were
	// just testing a url
	totalUrlCount++;
	
	if ([[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_ENABLE_FILTERING] == NO && _testMode == NO)
		return YES;

	NSString* referrerString = _referrer != nil ? [_referrer absoluteString] : nil;
	NSString* urlString = _url != nil ? [_url absoluteString] : nil;
	
	// need to check the referrer first so that we catch subresource loads properly
	// otherwise, it's fine to check the url itself
	if ([self shouldLoadWithoutFilter: (_referrer != nil) ? _referrer : _url] == YES)
		return YES;
	
	// if we are a subresource and cross-site content is blocked, do the domain checking
	// how fast in an NSUserDefaults lookup? fast i hope since we are going to do this alot
	if (_referrer != nil &&
			[[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_BLOCK_CROSS_SITE_CONTENT] == YES)
	{
		NSString* referringHost = [_referrer host];
		NSString* targetHost = [_url host];
		
		// fixme: is there a better/faster way to do this?
		// NSRange referringDomain = [referringHost rangeOfString:@"." options:(NSCaseInsensitiveSearch | NSLiteralSearch | NSBackwardsSearch)];
		NSArray* referringHostComponents = [referringHost componentsSeparatedByString:@"."];
		NSArray* targetHostComponents = [targetHost componentsSeparatedByString:@"."];
		
		// compare the last two components of each list - this should be the TLD and domain name
		if (![[referringHostComponents objectAtIndex:[referringHostComponents count] - 1] isEqualToString:[targetHostComponents objectAtIndex:[targetHostComponents count] - 1]] ||
		![[referringHostComponents objectAtIndex:[referringHostComponents count] - 2] isEqualToString:[targetHostComponents objectAtIndex:[targetHostComponents count] - 2]])
		{
			if ([_url isFileURL])
				return YES;

			blockedUrlCount++;
			*_logResults = [self logAction:PH_ACTION_BLOCK forUrl:urlString withReferrer:referringHost byRule:PH_RULE_CROSS_SITE withResult:nil testMode:_testMode];
			return NO;
		}
	}

	PithHelmetRuleNodeData* blockRule;
	NSEnumerator* blockEnumerator = [[self blockRules] objectEnumerator];
	
	while (blockRule = [blockEnumerator nextObject])
	{
		// skip subresource rules when there is no referrer
		if ([blockRule isSubresourceRule] == YES && _referrer == nil)
			continue;

		if ([blockRule matchURL:_url] == YES)
		{
			PithHelmetRuleNodeData* allowRule;
			NSEnumerator* includeEnumerator = [[self allowRules] objectEnumerator];
		
			// if we match an include pattern explicitly, continue loading
			while (allowRule = [includeEnumerator nextObject])
			{
				if ([allowRule matchURL:_url] == YES)
				{
					*_logResults = [self logAction:PH_ACTION_ALLOW forUrl:urlString withReferrer:referrerString byRule:[allowRule name] withResult:nil testMode:_testMode];
					allowedUrlCount++;
					return YES;
				}
			}
			*_logResults = [self logAction:PH_ACTION_BLOCK forUrl:urlString withReferrer:referrerString byRule:[blockRule name] withResult:nil testMode:_testMode];
			blockedUrlCount++;
			return NO;
		}
	}
	
	*_logResults = @"";
	return YES;
}

/**
 * Chooses a block file url to substitute for the given url.
 * Currently just returns one path, but might eventually filter by url
 * Creates a new url each time in case the other code does something funky to the object later on.
 */
- (NSURL*) blockedFileForUrl:(NSURL*)_url
{
	NSString* blockedFile = nil;	
	NSString* fileType = [[_url path] pathExtension];
	if ([fileType length] != 3 && [fileType length] != 4)
		fileType = @"";

	NSBundle* bundle = [NSBundle bundleWithIdentifier:PH_IDENTIFIER];
	blockedFile = [bundle pathForResource:@"blocked" ofType:fileType];
	if (blockedFile == nil)
		blockedFile = [bundle pathForResource:@"blocked" ofType:nil];

	DLog1(DLOG_30, (@"blockedFile(%@): %@", fileType, blockedFile));

	NSURL* blockedURL = [NSURL fileURLWithPath:blockedFile];
	DLog1(DLOG_20, (@"blockedURL:<%@> blockedFile:<%@>", blockedURL, blockedFile));
	return blockedURL;
}

- (NSString*) transmogrify:(NSURL*)_url
{
	NSString* logResults;
	return [self transmogrify:_url testMode:NO logResults:&logResults];
}

/**
 * Attempts a transmogrification.  If it fails, or no suitable rule is found, it
 * will return nil.
 */
- (NSString*) transmogrify:(NSURL*)_url testMode:(BOOL)_testMode logResults:(NSString**)_logResults
{
	if (_testMode == NO && ([[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_ENABLE_FILTERING] == NO ||
		[[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_ENABLE_TRANSMOGRIFIER] == NO))
		return nil;

	PithHelmetRuleRegexpReplace* rule;
	NSEnumerator* transEnumerator = [transmogrificationRules objectEnumerator];
	
	while (rule = [transEnumerator nextObject])
	{
		NSString* newUrl = [rule replace:_url];
		if (newUrl != nil)
		{
			*_logResults = [self logAction:PH_ACTION_TRANSMOGRIFY forUrl:[_url absoluteString] withReferrer:nil byRule:[rule description] withResult:newUrl testMode:_testMode];
			return newUrl;
		}
	}
	
	return nil;
}

- (NSString*) collapseContentScriptlet
{
	return collapseContentScriptlet;
}

/**
 * Creates a javascriptlet build from CollapseContent.js, substituting the current set of
 * regex rules.
 */
- (void) buildContentScriptlet
{
	NSString* javascriptPath = [[NSBundle bundleWithIdentifier:PH_IDENTIFIER] pathForResource:@"CollapseContent" ofType:@"js"];
	
	// if we had a previous value, release it
	if (collapseContentScriptlet != nil)
		[collapseContentScriptlet autorelease];
	
	collapseContentScriptlet = [NSMutableString stringWithContentsOfFile:javascriptPath];
	[collapseContentScriptlet retain];
	
	PithHelmetRuleNodeData* ruleData;
	NSEnumerator* e;

	NSMutableString* blockRegexpList = [NSMutableString stringWithCapacity:1024];
	e = [[self blockRules] objectEnumerator];
	while (ruleData = [e nextObject])
		if ([ruleData isValid] == YES)
			[blockRegexpList appendFormat:@"/%@/,", [ruleData rule]];
	
	NSMutableString* allowRegexpList = [NSMutableString stringWithCapacity:1024];
	e = [[self allowRules] objectEnumerator];
	while (ruleData = [e nextObject])
		if ([ruleData isValid] == YES)
			[allowRegexpList appendFormat:@"/%@/,", [ruleData rule]];

	NSString* collapseAdSizedImages = ([[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_COLLAPSE_AD_SIZED_IMAGES] == YES) ? @"true" : @"false";
	NSString* collapseBlockedLinks = ([[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_COLLAPSE_BLOCKED_LINKS] == YES) ? @"true" : @"false";

	[collapseContentScriptlet replaceOccurrencesOfString:@"${COLLAPSE_AD_SIZED_IMAGES}" withString:collapseAdSizedImages options:0 range:NSMakeRange(0, [collapseContentScriptlet length])];
	[collapseContentScriptlet replaceOccurrencesOfString:@"${COLLAPSE_BLOCKED_LINKS}" withString:collapseBlockedLinks options:0 range:NSMakeRange(0, [collapseContentScriptlet length])];
	[collapseContentScriptlet replaceOccurrencesOfString:@"${BLOCK_PATTERNS}" withString:blockRegexpList options:0 range:NSMakeRange(0, [collapseContentScriptlet length])];
	[collapseContentScriptlet replaceOccurrencesOfString:@"${ALLOW_PATTERNS}" withString:allowRegexpList options:0 range:NSMakeRange(0, [collapseContentScriptlet length])];

	DLog1(DLOG_30, (@"buildContentScriptlet: %@", collapseContentScriptlet));
}


/**
 * Reads the user defaults and sets un-set values to factory defaults.
 */
- (void) readUserDefaults
{
	NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
	totalUrlCount = [defaults integerForKey:PH_PREF_KEY_TOTAL_URL_COUNT];
	blockedUrlCount = [defaults integerForKey:PH_PREF_KEY_BLOCKED_URL_COUNT];
	allowedUrlCount = [defaults integerForKey:PH_PREF_KEY_ALLOWED_URL_COUNT];

	if ([defaults objectForKey:PH_PREF_KEY_COLLAPSE_AD_SIZED_IMAGES] == nil)
		[defaults setBool:YES forKey:PH_PREF_KEY_COLLAPSE_AD_SIZED_IMAGES];

	if ([defaults objectForKey:PH_PREF_KEY_ENABLE_CONTEXTUAL_MENU] == nil)
		[defaults setBool:YES forKey:PH_PREF_KEY_ENABLE_CONTEXTUAL_MENU];

	if ([defaults objectForKey:PH_PREF_KEY_COLLAPSE_BLOCKED_CONTENT] == nil)
		[defaults setBool:YES forKey:PH_PREF_KEY_COLLAPSE_BLOCKED_CONTENT];

	if ([defaults objectForKey:PH_PREF_KEY_COLLAPSE_BLOCKED_LINKS] == nil)
		[defaults setBool:YES forKey:PH_PREF_KEY_COLLAPSE_BLOCKED_LINKS];

	if ([defaults objectForKey:PH_PREF_KEY_ENABLE_FILTERING] == nil)
		[defaults setBool:YES forKey:PH_PREF_KEY_ENABLE_FILTERING];

	if ([defaults objectForKey:PH_PREF_KEY_PLAY_ANIMATED_GIFS] == nil)
		[defaults setObject:PH_PREF_VAL_PLAY_ANIMATED_GIFS_ALWAYS forKey:PH_PREF_KEY_PLAY_ANIMATED_GIFS];
}

- (void) saveUserDefaults
{
	NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
	[defaults setInteger:totalUrlCount forKey:PH_PREF_KEY_TOTAL_URL_COUNT];
	[defaults setInteger:blockedUrlCount forKey:PH_PREF_KEY_BLOCKED_URL_COUNT];
	[defaults setInteger:allowedUrlCount forKey:PH_PREF_KEY_ALLOWED_URL_COUNT];
	[defaults synchronize];
}

- (int) totalUrlCount
{
	return totalUrlCount;
}

- (int) allowedUrlCount
{
	return allowedUrlCount;
}

- (int) blockedUrlCount
{
	return blockedUrlCount;
}

- (float) blockedPercentage
{
	if (totalUrlCount == 0)
		return 0.0;
	
	return 100.0 * blockedUrlCount / totalUrlCount;
}

- (BOOL) shouldCollapseContentForUrl:(NSURL*)_url
{
	if ([self shouldLoadWithoutFilter:_url] == YES)
		return NO;
	
	return ([[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_ENABLE_FILTERING] == YES && [[NSUserDefaults standardUserDefaults] boolForKey:PH_PREF_KEY_COLLAPSE_BLOCKED_CONTENT] == YES);
}

- (void) setLoaded:(BOOL)_loaded
{
	loaded = _loaded;
}

- (BOOL) isLoaded
{
	return loaded;
}

- (NSString*) version
{
	return [[NSBundle bundleWithIdentifier:PH_IDENTIFIER] objectForInfoDictionaryKey: @"CFBundleShortVersionString"];
}

- (NSString*) nibName
{
	return [[NSBundle bundleWithIdentifier:PH_IDENTIFIER] objectForInfoDictionaryKey: @"NSMainNibFile"];
}


- (NSString*) bundleVersion
{
	return [[NSBundle bundleWithIdentifier:PH_IDENTIFIER] objectForInfoDictionaryKey: @"CFBundleVersion"];
}


- (void) setBlockRules:(NSMutableArray*)_blockRules
{
	if (blockRules != nil)
		[blockRules autorelease];
	
	blockRules = [_blockRules retain];
}

-(NSArray*) blockRules
{
	return blockRules;
}

- (void) setAllowRules:(NSMutableArray*)_allowRules;
{
	if (allowRules != nil)
		[allowRules autorelease];
	
	allowRules = [_allowRules retain];
}

-(NSArray*) allowRules
{
	return allowRules;
}

- (void) setExemptSiteRules:(NSMutableArray*)_rules;
{
	if (exemptSiteRules != nil)
		[exemptSiteRules autorelease];
	
	exemptSiteRules = [_rules retain];
}

-(NSArray*) exemptSiteRules
{
	return exemptSiteRules;
}

+ (void) errorAlert:(NSString*)_message withDetails:(NSString*)_details, ...
{
	va_list ap;
	va_start(ap, _details);

	NSString* detailsFormatted = [[[NSString alloc] initWithFormat:_details arguments:ap] autorelease];
	va_end(ap);

	NSBeginAlertSheet(
		_message,         // sheet message
		nil,              // default button label
		nil,              // alternate button label
		nil,              // no third button
		nil,              // window sheet is attached to
		self,             // we'll be our own delegate
		nil,              // did-end selector
		nil,              // no need for did-dismiss selector
		nil,              // context info
		detailsFormatted,	// additional text
		nil);             // no parameters in message
}

@end
